function [TIME,RES,RES_LEAK] = NUCLEAR_COOLING(NUCLEAR_REFRIGERANT,...
                                               INITIAL_HEAT_LEAK,...
                                               ISLAND_VOLUME,...
                                               FILENAME,...
                                               TRESHOLD,...
                                               HEAT_LEAK_INCREMENT,...
                                               RESIDUALS_LENGTH,...
                                               HEAT_CAPACITY_CUTOFF,...
                                               TEMPERATURE_CUTOFF)

%----------------------------------------------------------------------
%                          INITIALIZATION
%----------------------------------------------------------------------                                                     
                                                     
% Initialize physical constants                                                     
                                                     
[dataFilePath, dataFileName, ~] = fileparts('INIT.ini');
metaDataFile = fullfile(dataFilePath, [dataFileName '.ini']);
metadata = INI2STRUCT(metaDataFile);
   
mu_n  = str2double(metadata.constants.nucl_magneton);
k_B   = str2double(metadata.constants.boltz_const);
A     = str2double(metadata.constants.avogdr_const);
R     = k_B*A;

% Initialize material constants

[dataFilePath, dataFileName, ~] = fileparts('MAT.ini');
metaDataFile = fullfile(dataFilePath, [dataFileName '.ini']);
metadata = INI2STRUCT(metaDataFile);
   
if strcmp(NUCLEAR_REFRIGERANT,'in')==1

I     = str2double(metadata.in.nucl_spin);  
g_n   = str2double(metadata.in.nucl_g_factor);  
M     = str2double(metadata.in.molar_mass);   
kappa = str2double(metadata.in.korringa_const);
rho   = str2double(metadata.in.density);

% Nuclear heat capacity
n     = ISLAND_VOLUME*rho/M;

% Heat capacity of conduction electrons
alpha_e = str2double(metadata.in.heat_cap_alpha);
alpha_e = alpha_e*n;

Q = 3.1672e-26;
   
end

if strcmp(NUCLEAR_REFRIGERANT,'cu')==1

I     = str2double(metadata.cu.nucl_spin);  
g_n   = str2double(metadata.cu.nucl_g_factor);  
M     = str2double(metadata.cu.molar_mass);   
kappa = str2double(metadata.cu.korringa_const);
rho   = str2double(metadata.cu.density);

% Nuclear heat capacity
n     = ISLAND_VOLUME*rho/M; 

% Heat capacity of conduction electrons
alpha_e = str2double(metadata.cu.heat_cap_alpha);
alpha_e = alpha_e*n;

Q = 0;
   
end

%----------------------------------------------------------------------
%                     DYNAMIC SPIN TEMPERATURE
%----------------------------------------------------------------------

% Load the supporting datafile, the model needs the field profile in time
% as well as the measured electron temperature, it will calculate the 
% nuclear spin temperature and the heat leak from it then.

load(FILENAME);

% The stability is determined by the timescale of the spin-lattice
% relaxation and Delta_t, choose Delta_t short enough
Delta_t = 5e-6;
TIME_FACTOR = MEAS(length(MEAS(:,1)),1)/60;
time_steps = TIME_FACTOR*3.6e3/Delta_t;

% Starting temperature
T_START = MEAS(1,3);    
    
% Starting field, end field and field steps 
Bi = MEAS(1,2);
Bf = MEAS(length(MEAS(:,2)),2);  
Delta_B = (Bi-Bf)/time_steps;
    
% The size of the resulting temperature array
RES_LENGTH = length(MEAS(:,1));
RES=zeros(RES_LENGTH,3);     

% The timescale of the measured data
TIME = MEAS(:,1);

% Initialize the system
[Te,Tn] = deal(T_START);
B_FIELD = Bi;

% The resulting heat leak for every temperature point in RES
RES_LEAK = zeros(length(RES(:,1)),1);

% If a single value is given for the heat leak the model will calculate
% the corresponding heat leak per point, if an array is given the model
% will just run with the given heat leak

if length(INITIAL_HEAT_LEAK) == 1
    
    Q_leak = INITIAL_HEAT_LEAK;
    
end

% Main loop of the simulation

        for i=1:RES_LENGTH
        
        % The result of the current step is saved    
        RES(i,1:3) = [Tn Te B_FIELD];
        T_TEMP = [Tn Te];
        
        % The range over which the nuclear spin entropy for a small
        % change in field is calculated
         RANGE = 1e-4;
         T  = (Tn-RANGE):1e-8:(Tn+RANGE);
 
         % Calculate the nuclear spin eigenenergies at B1 and B2=B1-Delta_B
         E_1 = -(mu_n*g_n.*B_FIELD).*(-I:1:I)...
               -(Q/(4*I*(2*I-1)))...
               .*(3.*((-I:1:I)).^2-I*(I+1));
            
         E_2 = -(mu_n*g_n.*(B_FIELD-((time_steps/RES_LENGTH)*Delta_B)))...
               .*(-I:1:I)...
               -(Q/(4*I*(2*I-1)))...
               .*(3.*((-I:1:I)).^2-I*(I+1));

         % Calculate the partition sum at B1 and B2=B1-Delta_B
         Z_1 = 0; Z_2 = 0;
         
             for k=1:length(E_1)
                 
                 Z_1=Z_1+exp(-E_1(k)./(k_B*T)); 
                 Z_2=Z_2+exp(-E_2(k)./(k_B*T)); 
                 
             end
             
         Z_1 = T.*log(Z_1); Z_2 = T.*log(Z_2); 
         
         % Calculate the nuclear spin entropy and nuclear spin heat
         % capacity from the partition sum
             
         S_1 = n*(R)*DIFF_NUM(Z_1,T); 
         S_2 = n*(R)*DIFF_NUM(Z_2,T); 
         S_1 = S_1(2:length(S_1)); 
         S_2 = S_2(2:length(S_2)); 
         T = T(2:length(T));
         C_1 = T.*DIFF_NUM(S_1,T); 
         
         % Find the working point of the system in S(T)
         
         Tn_POS = find(abs(T-Tn)==min(abs(T-Tn))); 
         S_POS = S_1(Tn_POS);            
         S_2_POS = find(abs(S_2-S_POS)==min(abs(S_2-S_POS)));
            
         % Calculate the temperature change by adiabatically going from
         % B1 to B2=B1-Delta_B
         
         Delta_T = (T(Tn_POS(1))-T(S_2_POS(1)))/(time_steps/RES_LENGTH);
         
         % Catching roundoff errors if it happens
        
            for l=1:25
                
            C_1 = SMOOTH(C_1);
            
            end
            
         % The nuclear spin heat capacity    
                       
         Cn = C_1(Tn_POS); 
            
            if length(INITIAL_HEAT_LEAK) > 1
                
                Q_leak = INITIAL_HEAT_LEAK(i);
                
            end
            
            % Let the system go with the heat leak
            
            for j=1:time_steps/RES_LENGTH

                Tn =  Tn + ((Te*Tn-Tn^2)/kappa)*Delta_t...
                         - Delta_T;
    
                Te =  Te -(Cn/(alpha_e*Te))...
                         *((Tn*Te-Tn^2)/kappa)*Delta_t...
                         +(Q_leak/(alpha_e*Te))*Delta_t; 
                                 
            end
            
        % Now the heat leak is adapted in order to match the end point
        % of the step in S(T,B) space according to the measured electron
        % temperature. When the nuclear heat capacity Cn is depleted 
        % below a give value this becomes unstable and will not be performed
        
        while abs(Te-MEAS(i,3))>TRESHOLD...
              && i<RESIDUALS_LENGTH && Cn > HEAT_CAPACITY_CUTOFF
            
            disp(['STEP ' num2str(i)...
                  ', Tn = ' num2str(Tn)...
                  ', Te = ' num2str(Te)...
                  ', B = ' num2str(B_FIELD)...
                  ', C = ' num2str(Cn)...
                  ', Q = ' num2str(Q_leak)])
           
            Q_leak = Q_leak...
                     + sign(MEAS(i,3)-Te)...
                     *(INITIAL_HEAT_LEAK*HEAT_LEAK_INCREMENT);
                 
            Tn = T_TEMP(1); Te = T_TEMP(2);
            
                for j=1:time_steps/RES_LENGTH
        
                    Tn =  Tn + ((Te*Tn-Tn^2)/kappa)*Delta_t...
                             - (Tn/B_FIELD)*Delta_B;
    
                    Te =  Te -(Cn/(alpha_e*Te))...
                             *((Tn*Te-Tn^2)/kappa)*Delta_t...
                             +(Q_leak/(alpha_e*Te))*Delta_t; 
                                                     
                end
        
        end
        
        

            disp(['STEP ' num2str(i)...
                  ', Tn = ' num2str(Tn)...
                  ', Te = ' num2str(Te)...
                  ', B = ' num2str(B_FIELD)...
                  ', C = ' num2str(Cn)...
                  ', Q = ' num2str(Q_leak)])
        
        % Stop the simulation if the electron temperature explodes
              
                if Te > TEMPERATURE_CUTOFF                     
                    return;                     
                end
        
        RES_LEAK(i) = Q_leak;
        
        % The end field is the starting field for the next step
        
        B_FIELD = B_FIELD - ((time_steps/RES_LENGTH)*Delta_B);
        
        end
     
%----------------------------------------------------------------------
%                    END OF MAIN FUNCTION
%----------------------------------------------------------------------
        
end

%----------------------------------------------------------------------
%                    AUXILIARY FUNCTIONS
%----------------------------------------------------------------------

function [c] = DIFF_NUM(a,b)

pos=2;
c=zeros(1,length(a)-2);

    for i=2:length(b)
    
        c(pos)=(a(i)-a(i-1))/(b(i)-b(i-1));  
        pos=pos+1; 
  
    end

end

function [B] = SMOOTH(A)

B=zeros(1,length(A));

    for i=2:length(A)-1  
      
        B(i)=(1/3)*...
            (A(i-1)...
            +A(i)...
            +A(i+1));          
    end
 
B(1) = A(1);
B(length(B))=A(length(A));

end

function Result = INI2STRUCT(filename)

Result        = [];
currMainField = '';
f             = fopen(filename, 'r');

while ~feof(f)
    
    s = strtrim(fgetl(f));
    
    if isempty(s)
        continue;
    end;
    
    if s(1)==';'
        continue;
    end;
    
    if s(1)=='#'
        continue;
    end;
    
    if ( s(1)=='[' ) && ( s(end)==']' )
        currMainField = genvarname(lower(s(2:end-1)));
        Result.(currMainField) = [];    
    else
        [par,val] = strtok(s, '=');
        val = cleanValue(val);
        if ~isempty(currMainField)
            Result.(currMainField).(lower(genvarname(par))) = val;
        else
            Result.(lower(genvarname(par))) = val;
        end
    end
end

fclose(f);

    function [res] = cleanValue(s)
        res = strtrim(s);
        if strcmpi(res(1), '=')
            res(1) = [];
            res = strtrim(res);
        end
    end

end