import numpy as np
import scipy, scipy.constants
e = scipy.constants.e
k_B = scipy.constants.Boltzmann
import cbt_tools1

def blur_function(f,w,samples):
    x = np.cos( np.linspace(0,np.pi,samples) )*w
    def r(x_in, *args,**kwargs):
        x_out = []
        for x_el in x_in:
            x_out.append( np.mean( f(x_el + x,*args,**kwargs)) )
        return np.array(x_out)
    return r



def fit_single_curve(V,G,N,T_guess = 20e-3, C_sigma_guess = 300e-15, excitation = None, full_simulation = False):
    Gt_guess = np.max(G); V_offset_guess = 0.00001
    p0 = [T_guess,Gt_guess,C_sigma_guess,V_offset_guess]
    def G_first_fit(V,T,Gt,C_sigma,V_offset):
        return cbt_tools1.G_first_order(V,T,Gt,N,C_sigma,V_offset)
    if excitation is not None:
        G_first_fit = blur_function(G_first_fit,excitation,10)
    popt1, estimated_covariance = scipy.optimize.curve_fit(G_first_fit, V, G,p0 = p0)
    #print(popt1)
    def G_third_fit(V,T,Gt,C_sigma,V_offset):
        return cbt_tools1.G_third_order(V,T,Gt,N,C_sigma,V_offset)
    if excitation is not None:
        G_third_fit = blur_function(G_third_fit,excitation,10)
    popt2, estimated_covariance = scipy.optimize.curve_fit(G_third_fit, V, G,p0 = popt1)
    #print(popt2)
    V_offset = popt2[-1]
    estimated_std_dev2 = np.sqrt(np.diag(estimated_covariance))
    r = {'T':popt2[0],'Gt':popt2[1],'C_sigma':popt2[2],'V_offset':popt2[3]}
    errors = {'T':estimated_std_dev2[0],'Gt':estimated_std_dev2[1],'C_sigma':estimated_std_dev2[2],'V_offset':estimated_std_dev2[3]}
    fit = G_third_fit(V,*popt2)
    if full_simulation:
        def G_full_fit(V,T,Gt,C_sigma):
            return cbt_tools1.G_simulation(V,T,Gt,N,C_sigma,V_offset)
        if excitation is not None:
            G_full_fit = blur_function(G_full_fit,excitation,10)
        popt3, estimated_covariance = scipy.optimize.curve_fit(G_full_fit, V, G,p0 = popt2[:-1])
        print(popt3)
        estimated_std_dev3 = np.sqrt(np.diag(estimated_covariance))
        r = {'T':popt3[0],'Gt':popt3[1],'C_sigma':popt3[2],'V_offset':popt2[3]}
        errors = {'T':estimated_std_dev3[0],'Gt':estimated_std_dev3[1],'C_sigma':estimated_std_dev3[2],'V_offset':estimated_std_dev2[3]}
        fit = G_full_fit(V,*popt3)
    return r,errors,fit
  


def fit_multi_cbt(curves,N,T_guess = 20e-3, C_sigma_guess = 300e-15, excitation = None, full_simulation = False,verbose = True):
    #first fit all curves on their own.
    def p(*params,**kwargs):
        if verbose: print(*params,**kwargs)
    p('individual fits')
    for curve in curves:
        popt, stderror, fit = fit_single_curve(curve['V'],curve['G'],N,T_guess, C_sigma_guess, excitation = excitation, full_simulation = full_simulation)
        curve['single popt'] = popt
        curve['single stderror'] = stderror
        curve['single fit'] = fit
        p(popt['T'])
    p('building global parameters')
    #global parameters across the fits
    Gt = np.mean([c['single popt']['Gt'] for c in curves])
    C_sigma = np.mean([c['single popt']['C_sigma'] for c in curves])
    p0=[Gt,C_sigma]
    allG = np.array([])
    
    if excitation is not None:
        G_third_order = blur_function(cbt_tools1.G_third_order,excitation,6)
    else:
        G_third_order = cbt_tools1.G_third_order
            
    #make the unique parameters for each, T, V_offset, 
    for curve in curves:
        p0.append( curve['single popt']['T'] )
        p0.append( curve['single popt']['V_offset'] )
        allG=np.append(allG,curve['G'])

    Vparts = [c['V'] for c in curves]
    p('starting multifit')
    def _G_multi_fit(Vparts,Gt,C_sigma,*params):
        r = np.array([])
        for i,V in enumerate(Vparts):
            T,V_offset = params[2*i],params[2*i+1]
            r = np.append(r,  G_third_order(V,T,Gt,N,C_sigma,V_offset) )
        return r
    
    popt, pcov = scipy.optimize.curve_fit(_G_multi_fit,Vparts, allG , p0=p0)
    estimated_std_dev = np.sqrt(np.diag(pcov))
    #go a put the temperature and offset into each curve from the multifit.
    
    for i,curve in enumerate(curves):
        curve['popt'] = {
            'T' : popt[2 + i*2],
            'V_offset' : popt[2 + i*2 + 1],
            'Gt' : popt[0],
            'C_sigma' : popt[1]
        }
        curve['stderror'] = {
            'T' : estimated_std_dev[2 + i*2],
            'V_offset' : estimated_std_dev[2 + i*2 + 1],
            'Gt' : estimated_std_dev[0],
            'C_sigma' : estimated_std_dev[1]
        }
        
        curve['fit'] =  G_third_order(curve['V'],N=N,**curve['popt'])
        
    p('done')
    return curves,popt
